@@ -1,14 +1,28 @@
+"""
+`Dataset`s and `LayerView`s form the core abstractions in mundipy.
+
+`Dataset` comprises any source for vector data. Instantiating a
+`Dataset` declares its accessibility, but does not automatically
+load features, as all features are lazily loaded.
+
+`LayerView` represents a collection of vector features, typically
+a subset from a `Dataset`. This makes queries like intersection
+and nearest much faster because only a subset of the `Dataset`
+must be loaded.
+"""
+
 from shapely.geometry.base import BaseGeometry
 from shapely.geometry import box, Point, Polygon, MultiPolygon
 import shapely.wkt
+from shapely.ops import transform
 import geopandas as gpd
 import pandas as pd
 from sqlalchemy import create_engine
 from functools import lru_cache, partial
 import s2sphere
 
-from mundipy.cache import spatial_cache_footprint
-from mundipy.geometry import from_dataframe
+from mundipy.cache import spatial_cache_footprint, pyproj_transform
+from mundipy.geometry import from_dataframe, from_row_series
 
 r = s2sphere.RegionCoverer()
 
@@ -23,41 +37,46 @@ def tile_bbox(polygon):
 	# 14 is ~0.3km2 and 600m edge length
 	return r.get_simple_covering(rect, p1.to_point(), 14)
 
-"""A Layer represents a group of spatial data."""
-class Layer:
+class Dataset:
+	"""
+	A Dataset represents a source of vector features.
+
+	from mundipy.layer import Dataset
+
+	src = Dataset({
+ 		'url': 'postgresql://postgres@localhost:5432/postgres',
+		'table': 'table_name'
+	})
+
+	"""
 
 	def __init__(self, data):
+		""" Initialize a Dataset from a data source. """
+
 		self.filename = None
-		self._dataframe = None
 		self._db_url = None
 		self._db_table = None
 
 		self._conn = None
 
-		""" Initialize a Layer from a data source. """
-		if isinstance(data, gpd.GeoDataFrame):
-			self._dataframe = data
-		elif isinstance(data, dict):
+		if isinstance(data, dict):
 			self._db_url = data['url']
 			self._db_table = data['table']
 		elif isinstance(data, str):
 			self.filename = data
 		else:
-			raise TypeError('data for Layer() is not filename, GeoDataFrame or dict with PostgreSQL details')
+			raise TypeError('data for Dataset() is neither filename nor dict with PostgreSQL details')
 
 	@spatial_cache_footprint
 	def _load(self, bbox, pcs='EPSG:4326'):
 		"""
-		Load part or the entire Layer as a dataframe.
+		Load part or the entire Dataset as a dataframe.
 
 		Takes bbox as a 4-tuple of WGS84 coordinates (lon, lat). bbox can be
 		None to load the entire dataset.
 
 		Returns the dataset in PCS coordinates.
 		"""
-		if self._dataframe is not None:
-			return (self._dataframe, None)
-
 		if self._db_url is not None:
 			# load from PostGIS
 			self._conn = create_engine(self._db_url)
@@ -111,7 +130,7 @@ class Layer:
 
 	@property
 	def dataframe(self):
-		"""Load an entire Layer as a dataframe."""
+		"""Load an entire Dataset as a dataframe."""
 		return self._load(None)
 
 	@lru_cache(maxsize=8)
@@ -122,52 +141,83 @@ class Layer:
 	def geometry_collection(self, pcs):
 		return from_dataframe(self.local_dataframe(pcs))
 
-	"""Read into a Layer at a specific geometry (WGS84)."""
+	"""Read into a Dataset at a specific geometry (WGS84)."""
 	def inside_bbox(self, bbox, pcs='EPSG:4326'):
 		if not isinstance(bbox, tuple) or len(bbox) != 4:
 			raise TypeError('inside_bbox expected bbox to be a 4-tuple')
 
 		return self._load(box(*bbox), pcs=pcs)
 
+class LayerView:
+	"""
+	`LayerView` represents a collection of vector features in
+	a single dataset. It implements an `Iterable` interface,
+	allowing for one to loop through all features in the dataset,
+	or smart filtering without loading the entire dataset into
+	memory.
+	"""
 
-"""VisibleLayer represents Layer data, as seen from a geometry."""
-class VisibleLayer:
-	# Center needs to be projected to the local CRS already
-	def __init__(self, layer, bbox, center, pcs=None):
-		# do local calculations in a projected CRS, which is 2D and has
-		# typical distance units
-		if not isinstance(pcs, str):
-			raise TypeError('pcs passed to VisibleLayer() was not a string')
+	def __init__(self, layer, pcs):
+		self.layer = layer
 		self.pcs = pcs
 
-		if not isinstance(layer, Layer):
-			raise TypeError('mapdata passed to VisibleLayer() was not a mundipy.Map')
-		# results come as projected coordinate system
-		self.local_collection = layer.inside_bbox(bbox, pcs=pcs)
+	def __iter__(self):
+		"""
+		Iterate through all items of the dataset.
+		"""
+		yield from self.layer.geometry_collection(self.pcs)
 
-		if not isinstance(center, BaseGeometry):
-			raise TypeError('center passed to VisibleLayer was not shapely.geometry.BaseGeometry')
-		self.center = center
+	def intersects(self, geom):
+		"""
+		Returns an `Iterator` of mundipy geometries that intersect
+		with `geom`.
 
-	def intersects(self):
-		"""Give a list of members that intersect with the center geometry."""
-		possible_matches_index = list(self.local_collection.sindex.intersection(self.center.bounds))
+		`geom` - inherits from `shapely.geometry`
 
-		# indexing into dataframes is expensive, compute a new list of actual matches
-		# .iloc only takes integer locations, even for columns
-		geometry_col_loc = self.local_collection.columns.get_loc('geometry')
-		actual_matches_index = [match for match in possible_matches_index if self.local_collection.iloc[match, geometry_col_loc].intersects(self.center)]
+        from mundipy.utils import plot
 
-		return self.local_collection.iloc[actual_matches_index]
+        for feat in layer.intersects(Point(-37.0, 42.1)):
+            plot(feat)
+		"""
+		if not isinstance(self.layer, Dataset):
+			raise TypeError('intersects() on not Dataset undefined')
+		if not isinstance(geom, BaseGeometry):
+			raise TypeError('geom is not a shapely.geometry')
 
-	def nearest(self):
-		"""Give the closest member to the center geometry."""
-		spatial_index = self.local_collection.sindex
+		# convert geom to EPSG:4326
+		to_wgs = pyproj_transform(self.pcs, 'EPSG:4326')
+		# buffer a little bit to prevent point
+		bbox = transform(to_wgs, geom.buffer(1e-8)).bounds
 
-		idx = spatial_index.nearest(self.center.centroid)
-		return self.local_collection.iloc[idx[1][0]]
+		potentially_intersecting_gdf = self.layer.inside_bbox(bbox, self.pcs)
+		return from_dataframe(potentially_intersecting_gdf[potentially_intersecting_gdf.intersects(geom)])
 
-	def buffer(self, distance):
-		"""Buffer the geometry by some distance."""
+	def nearest(self, geom):
+		"""
+		Returns the nearest feature in this collection to the passed
+		geometry.
 
-		return self.local_collection.geometry.buffer(distance)
+		Returns `None` if the dataset has no features.
+
+		`geom`: inherits from `shapely.geometry`
+		"""
+		if not isinstance(self.layer, Dataset):
+			raise TypeError('intersects() on not Dataset undefined')
+		if not isinstance(geom, BaseGeometry):
+			raise TypeError('geom is not a shapely.geometry')
+
+		to_wgs = pyproj_transform(self.pcs, 'EPSG:4326')
+
+		# increasing look outside of this bbox for the nearest item
+		buffer_distances = [1e0, 1e1, 1e2, 1e3, 1e4, 1e8]
+		for buffer_size in buffer_distances:
+			# buffer geom.bbox
+			bbox = transform(to_wgs, geom.buffer(buffer_size)).bounds
+
+			gdf = self.layer.inside_bbox(bbox, self.pcs)
+			if len(gdf) > 0:
+				res = min(gdf.iterrows(), key=lambda row: geom.distance(row[1].geometry))
+
+				return from_row_series(res[1])
+
+		return None
