@@ -2,7 +2,9 @@ import json
 import urllib.parse
 import difflib
 import inspect
+from contextvars import copy_context
 
+import fiona
 from tqdm import tqdm
 import geopandas as gpd
 import pandas as pd
@@ -14,11 +16,12 @@ import matplotlib.colors as mcolors
 import matplotlib.patches as mpatches
 
 from mundipy.map import Map
-from mundipy.layer import Layer, VisibleLayer
+from mundipy.layer import Dataset, LayerView
 from mundipy.api.osm import grab_from_osm
 from mundipy.pcs import choose_pcs, NoProjectionFoundError
 from mundipy.cache import pyproj_transform
 from mundipy.geometry import from_row_series
+from mundipy.utils import _plot
 
 class MundiQ:
     def __init__(self, center, mapdata, plot_target=None, units='meters', clip_distance=500):
@@ -44,54 +47,24 @@ class MundiQ:
             raise TypeError('clip_distance to mundi.plot() must be float or int')
         self.clip_distance = clip_distance
 
-    """ Returns a VisibleLayer. """
-    def __call__(self, *args):
-        # variable length arguments
-        dataset = args[0] if len(args) > 0 else None
-
-        # First, if it's just Q(), give the center.
-        if dataset == None:
-            return self.center
-
-        # Next, if dataset is in the pool, return the
-        # VisibleLayer
-        if dataset in self.mapdata.collections.keys():
-            return VisibleLayer(self.mapdata.collections[dataset], self.bbox(), self.center.geometry, pcs=self.pcs)
-
-        if dataset == 'openstreetmap' or dataset == 'osm':
-            # Take either a tuple, or a list of tuples
-            if len(args) <= 1:
-                raise TypeError('openstreetmap Q() takes at least two arguments')
-
-            # build bounding box in WGS84
-            p1, p2, p3, p4 = self.bbox()
-
-            gdf = grab_from_osm(tups=args[1], bbox=('%f,%f,%f,%f' % (p2, p1, p4, p3)))
-            # convert to local CRS, because grab_from_osm gives WGS84
-            return VisibleLayer(Layer(gdf), self.bbox(), self.center.geometry, pcs=self.pcs)
-
-        # this is an unknown dataset, throw a useful error
-        possible_datasets = list(self.mapdata.collections.keys()) + ['openstreetmap', 'osm']
-        similar_dataset_name = difflib.get_close_matches(dataset, possible_datasets, n=1)
-
-        if len(similar_dataset_name) == 0:
-            raise TypeError('Unknown dataset name %s was passed to Q()' % dataset)
-        else:
-            raise TypeError('Unknown dataset name %s was passed to Q(), did you mean %s?' % (dataset, possible_datasets[0]))
-
     def call_process(self, fn):
         # pass dataset as dataframe if requested
-        args = inspect.getfullargspec(fn)[0][2:]
+        args = inspect.getfullargspec(fn)[0][1:]
 
         df_args = []
         for arg in args:
             try:
-                df_args.append(self.mapdata.collections[arg].geometry_collection(self.pcs))
+                df_args.append(LayerView(self.mapdata.collections[arg], self.pcs))
             except KeyError:
                 raise TypeError('mundi process() function requests dataset \'%s\', but no dataset was defined on Mundi' % arg)
 
         center = from_row_series(self.center)
-        return fn(self, center, *df_args)
+
+        # call fn with a relevant context
+        ctx = copy_context()
+        ctx.run(lambda: _plot.set(self.plot))
+
+        return ctx.run(fn, center, *df_args)
 
     def _bbox(self, distance=500):
         """Builds a bounding box around the center object in the local coordinate system."""
@@ -109,7 +82,7 @@ class MundiQ:
 
         # coerce shape into a GeoSeries, no matter what it is
         # shape can literally be anything
-        if isinstance(shape, Polygon) or isinstance(shape, MultiPolygon):
+        if isinstance(shape, Point) or isinstance(shape, Polygon) or isinstance(shape, MultiPolygon) or isinstance(shape, LineString):
             shape = gpd.GeoSeries([shape])
         elif isinstance(shape, gpd.GeoDataFrame):
             shape = shape.geometry
@@ -117,8 +90,6 @@ class MundiQ:
             shape = gpd.GeoSeries([shape.geometry])
         elif isinstance(shape, pd.Series) and isinstance(shape.geometry, gpd.GeoSeries):
             shape = shape.geometry
-        elif isinstance(shape, VisibleLayer):
-            shape = shape.local_collection
         else:
             raise TypeError('unexpected type passed to plot(), got %s' % type(shape))
 
@@ -131,7 +102,8 @@ class MundiQ:
             shape = gpd.clip(shape, mask=self._bbox(distance=self.clip_distance))
 
         # LineString and Point won't show up when plotted; buffer
-        shape = shape.apply(lambda g: g.buffer(2) if isinstance(g, LineString) or isinstance(g, Point) else g)
+        # 10 feet and meters should be, fine, i guess
+        shape = shape.apply(lambda g: g.buffer(10) if isinstance(g, LineString) or isinstance(g, Point) else g)
 
         # convert to WGS84
         shape = shape.set_crs(crs=self.pcs).to_crs(epsg=4326)
